# -*- coding: utf-8 -*-

"""
@date: 2020/12/2 下午9:38
@file: mobilenetv1_backbone.py
@author: zj
@description: 
"""

from abc import ABC
import torch.nn as nn

from zcls.model import registry
from zcls.model.init_helper import init_weights
from zcls.model.conv_helper import get_conv
from zcls.model.norm_helper import get_norm
from zcls.model.act_helper import get_act
from ..misc import make_divisible
from .mobilenetv1_block import MobileNetV1Block


def make_stem(in_planes,
              base_planes,
              conv_layer,
              norm_layer,
              act_layer
              ):
    return nn.Sequential(
        conv_layer(in_planes, base_planes, kernel_size=3, stride=2, padding=1, bias=False),
        norm_layer(base_planes),
        act_layer(inplace=True)
    )


class MobileNetV1Backbone(nn.Module, ABC):

    def __init__(self,
                 in_channels=3,
                 base_channels=32,
                 layer_channels=(64, 128, 128, 256, 256, 512, 512, 512, 512, 512, 512, 1024, 1024),
                 strides=(1, 2, 1, 2, 1, 2, 1, 1, 1, 1, 1, 2, 2),
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None
                 ):
        """
        :param in_channels: 输入通道数
        :param base_channels: 第一个卷积层通道数
        :param layer_channels: 后续深度卷积层通道数
        :param strides: 卷积步长
        :param conv_layer: 卷积层类型
        :param norm_layer: 归一化层类型
        :param act_layer: 激活层类型
        """
        super(MobileNetV1Backbone, self).__init__()
        assert len(strides) == len(layer_channels)

        if conv_layer is None:
            conv_layer = nn.Conv2d
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if act_layer is None:
            act_layer = nn.ReLU

        self.stem = make_stem(in_channels,
                              base_channels,
                              conv_layer,
                              norm_layer,
                              act_layer)
        self.layer_num = self._make_dw(base_channels,
                                       layer_channels,
                                       strides,
                                       conv_layer,
                                       norm_layer,
                                       act_layer)
        init_weights(self.modules())

    def _make_dw(self,
                 base_planes,
                 layer_planes,
                 strides,
                 conv_layer,
                 norm_layer,
                 act_layer
                 ):
        layer_num = len(layer_planes)

        in_planes = base_planes
        padding = 1
        for i, (layer_plane, stride) in enumerate(zip(layer_planes, strides)):
            if i == (layer_num - 1):
                # 最后一次深度卷积操作不进行空间尺寸下采样
                padding = 4
            dw_layer = MobileNetV1Block(int(in_planes),
                                        int(layer_plane),
                                        stride=stride,
                                        padding=padding,
                                        conv_layer=conv_layer,
                                        norm_layer=norm_layer,
                                        act_layer=act_layer
                                        )
            in_planes = layer_plane
            layer_name = f'layer{i + 1}'
            self.add_module(layer_name, dw_layer)

        return layer_num

    def _forward_impl(self, x):
        x = self.stem(x)

        for i in range(self.layer_num):
            layer = self.__getattr__(f'layer{i + 1}')
            x = layer(x)

        return x

    def forward(self, x):
        return self._forward_impl(x)


@registry.Backbone.register('MobileNetV1')
def build_mbv1_backbone(cfg):
    in_planes = cfg.MODEL.BACKBONE.IN_PLANES
    base_planes = cfg.MODEL.BACKBONE.BASE_PLANES
    layer_planes = cfg.MODEL.BACKBONE.LAYER_PLANES
    strides = cfg.MODEL.BACKBONE.STRIDES
    conv_layer = get_conv(cfg)
    norm_layer = get_norm(cfg)
    act_layer = get_act(cfg)
    width_multiplier = cfg.MODEL.COMPRESSION.WIDTH_MULTIPLIER
    round_nearest = cfg.MODEL.COMPRESSION.ROUND_NEAREST

    base_planes = make_divisible(base_planes * width_multiplier, round_nearest)
    layer_planes = [make_divisible(layer_plane * width_multiplier, round_nearest) for layer_plane in layer_planes]

    return MobileNetV1Backbone(
        in_channels=in_planes,
        base_channels=base_planes,
        layer_channels=layer_planes,
        strides=strides,
        conv_layer=conv_layer,
        norm_layer=norm_layer,
        act_layer=act_layer
    )



# -*- coding: utf-8 -*-

"""
@date: 2020/11/25 下午6:50
@file: dataloader.py
@author: zj
@description: 
"""

from yacs.config import CfgNode as CN


def add_config(_C):
    # ---------------------------------------------------------------------------- #
    # DataLoader
    # ---------------------------------------------------------------------------- #
    _C.DATALOADER = CN()
    # batch size per GPU
    _C.DATALOADER.TRAIN_BATCH_SIZE = 16
    _C.DATALOADER.TEST_BATCH_SIZE = 16

    # refert to [torch Dataloader中的num_workers](https://zhuanlan.zhihu.com/p/69250939)
    _C.DATALOADER.NUM_WORKERS = 2

    # random sample or sequential sample in train/test stage, default False
    _C.DATALOADER.RANDOM_SAMPLE = False

    # overlapped prefetching cpu->gpu memory copy, default False
    _C.DATALOADER.PREFETCHER = False


# -*- coding: utf-8 -*-

"""
@date: 2020/12/24 下午7:38
@file: shufflenetv1.py
@author: zj
@description: 
"""

from zcls.model import registry
from ..base_recognizer import BaseRecognizer

"""
G1) Equal channel width minimizes memory access cost (MAC)
G2) Excessive group convolution increases MAC.
G3) Network fragmentation reduces degree of parallelism
G4) Element-wise operations are non-negligible.
"""


@registry.RECOGNIZER.register('ShuffleNetV2')
class ShuffleNetV2(BaseRecognizer):

    def __init__(self, cfg):
        super().__init__(cfg)



# -*- coding: utf-8 -*-

"""
@date: 2020/12/3 下午8:52
@file: mobilenetv2_backbone.py
@author: zj
@description: 
"""

from abc import ABC
import torch.nn as nn

from zcls.model import registry
from zcls.model.init_helper import init_weights
from zcls.model.conv_helper import get_conv
from zcls.model.norm_helper import get_norm
from zcls.model.act_helper import get_act
from ..misc import make_divisible
from .mobilenetv2_block import MobileNetV2Block


def make_stem(in_planes,
              out_planes,
              kernel_size,
              stride,
              conv_layer,
              norm_layer,
              act_layer
              ):
    padding = 1 if stride == 2 else 0
    return nn.Sequential(
        conv_layer(in_planes, out_planes, kernel_size=kernel_size, stride=stride, padding=padding, bias=False),
        norm_layer(out_planes),
        act_layer(inplace=True)
    )


class MobileNetV2Backbone(nn.Module, ABC):

    def __init__(self,
                 in_channels=3,
                 out_channels=1280,
                 base_channels=32,
                 stride=2,
                 padding=1,
                 inverted_residual_setting=None,
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None,
                 ):
        """
        :param in_channels: 输入通道数
        :param out_channels: 输出通道数
        :param base_channels: 第一个卷积层通道数
        :param stride: 第一个卷积层步长
        :param padding: 第一个卷积层零填充
        :param inverted_residual_setting: 反向残差块设置
        :param conv_layer: 卷积层类型
        :param norm_layer: 归一化层类型
        :param act_layer: 激活层类型
        """
        super(MobileNetV2Backbone, self).__init__()

        if inverted_residual_setting is None:
            inverted_residual_setting = [
                # t, c, n, s
                [1, 16, 1, 1],
                [6, 24, 2, 2],
                [6, 32, 3, 2],
                [6, 64, 4, 2],
                [6, 96, 3, 1],
                [6, 160, 3, 2],
                [6, 320, 1, 1],
            ]
        else:
            assert len(inverted_residual_setting[0]) == 4

        if conv_layer is None:
            conv_layer = nn.Conv2d
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if act_layer is None:
            act_layer = nn.ReLU6

        kernel_size = 3
        stride = 2
        self.first_stem = make_stem(in_channels,
                                    base_channels,
                                    kernel_size,
                                    stride,
                                    conv_layer,
                                    norm_layer,
                                    act_layer
                                    )

        in_channels = base_channels
        for i, (t, c, n, s) in enumerate(inverted_residual_setting):
            feature_dims = c
            layer = MobileNetV2Block(in_channels,
                                     feature_dims,
                                     expansion_rate=t,
                                     repeat=n,
                                     stride=s,
                                     conv_layer=conv_layer,
                                     norm_layer=norm_layer,
                                     act_layer=act_layer)
            layer_name = f'layer{i + 1}'
            self.add_module(layer_name, layer)
            in_channels = feature_dims
        kernel_size = 1
        stride = 1
        self.last_stem = make_stem(in_channels,
                                   out_channels,
                                   kernel_size,
                                   stride,
                                   conv_layer,
                                   norm_layer,
                                   act_layer)

        self.layer_num = len(inverted_residual_setting)

        init_weights(self.modules())

    def forward(self, x):
        x = self.first_stem(x)

        for i in range(self.layer_num):
            layer = self.__getattr__(f'layer{i + 1}')
            x = layer(x)

        x = self.last_stem(x)
        return x


@registry.Backbone.register('MobileNetV2')
def build_mbv2_backbone(cfg):
    in_channels = cfg.MODEL.BACKBONE.IN_PLANES
    base_channels = cfg.MODEL.BACKBONE.BASE_PLANES
    out_channels = cfg.MODEL.HEAD.FEATURE_DIMS
    round_nearest = cfg.MODEL.COMPRESSION.ROUND_NEAREST
    width_multiplier = cfg.MODEL.COMPRESSION.WIDTH_MULTIPLIER
    conv_layer = get_conv(cfg)
    norm_layer = get_norm(cfg)
    act_layer = get_act(cfg)

    inverted_residual_setting = [
        # t, c, n, s
        [1, 16, 1, 1],
        [6, 24, 2, 2],
        [6, 32, 3, 2],
        [6, 64, 4, 2],
        [6, 96, 3, 1],
        [6, 160, 3, 2],
        [6, 320, 1, 1],
    ]

    base_channels = make_divisible(base_channels * width_multiplier, round_nearest)
    for i in range(len(inverted_residual_setting)):
        channel = inverted_residual_setting[i][1]
        inverted_residual_setting[i][1] = make_divisible(channel * width_multiplier, round_nearest)
    # out_channels = make_divisible(out_channels * width_multiplier, round_nearest)

    return MobileNetV2Backbone(
        in_channels=in_channels,
        out_channels=out_channels,
        base_channels=base_channels,
        inverted_residual_setting=inverted_residual_setting,
        conv_layer=conv_layer,
        norm_layer=norm_layer,
        act_layer=act_layer
    )



# -*- coding: utf-8 -*-

"""
@date: 2020/12/30 下午4:28
@file: mnasnet_unit.py
@author: zj
@description:
"""

from abc import ABC
import copy
import torch.nn as nn

from zcls.model import registry
from zcls.model.conv_helper import get_conv
from zcls.model.norm_helper import get_norm
from zcls.model.act_helper import get_act
from ..misc import round_to_multiple_of
from .mnasnet_unit import MNASNetUint, BN_MOMENTUM

arch_settings = {
    'mnasnet_a1': [
        # kernel_size, stride, expansion_rate, with_attention_2, repeated, out_planes
        [3, 1, 1, 0, 1, 16],
        [3, 2, 6, 0, 2, 24],
        [5, 2, 3, 1, 3, 40],
        [3, 2, 6, 0, 4, 80],
        [3, 1, 6, 1, 2, 112],
        [5, 2, 6, 1, 3, 160],
        [3, 1, 6, 0, 1, 320],
    ],
    'mnasnet_b1': [
        # refer to [vision/torchvision/models/mnasnet.py](https://github.com/pytorch/vision/blob/master/torchvision/models/mnasnet.py)
        # kernel_size, stride, expansion_rate, with_attention_2, repeated, out_planes
        [3, 1, 1, 0, 1, 16],
        [3, 2, 3, 0, 3, 24],
        [5, 2, 3, 0, 3, 40],
        [5, 2, 6, 0, 3, 80],
        [3, 1, 6, 0, 2, 96],
        [5, 2, 6, 0, 4, 192],
        [3, 1, 6, 0, 1, 320],
    ]
}


def make_stem(in_channels,
              base_channels,
              inner_channels,
              out_channels,
              conv_layer,
              norm_layer,
              act_layer):
    first_stem = nn.Sequential(
        conv_layer(in_channels, base_channels, kernel_size=3, stride=2, padding=1, bias=False),
        norm_layer(base_channels, momentum=BN_MOMENTUM),
        act_layer(inplace=True)
    )

    last_stem = nn.Sequential(
        conv_layer(inner_channels, out_channels, kernel_size=1, stride=1, padding=0, bias=False),
        norm_layer(out_channels, momentum=BN_MOMENTUM),
        act_layer(inplace=True)
    )

    return first_stem, last_stem


class MNASNetBackbone(nn.Module, ABC):

    def __init__(self,
                 in_channels=3,
                 out_channels=1280,
                 stage_setting=None,
                 width_multiplier=1.,
                 round_nearest=8,
                 with_attention=True,
                 reduction=4,
                 attention_type='SqueezeAndExcitationBlock2D',
                 conv_layer=None,
                 norm_layer=None,
                 act_layer=None
                 ):
        """
        MnasNet-A1实现
        参考Torchvision mnasnet实现，在Backbone之后添加一个Conv1x1-BN-ReLU，最终输出特征维度为1280
        :param in_channels: 输入通道数
        :param out_channels: 输出通道数
        :param stage_setting: 参数设置
        :param width_multiplier: 宽度乘法器
        :param round_nearest: 设置每一层通道数均为8的倍数
        :param with_attention: 是否使用注意力模块
        :param reduction: 衰减率
        :param attention_type: 注意力模块类型
        :param conv_layer: 卷积层类型
        :param norm_layer: 归一化层类型
        :param act_layer: 激活层类型
        """
        super(MNASNetBackbone, self).__init__()

        if conv_layer is None:
            conv_layer = nn.Conv2d
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if act_layer is None:
            act_layer = nn.ReLU
        block_layer = MNASNetUint

        base_planes = 32
        if stage_setting is None:
            stage_setting = [
                # kernel_size, stride, expansion_rate, with_attention_2, repeated, out_planes
                [3, 1, 1, 0, 1, 16],
                [3, 2, 6, 0, 2, 24],
                [5, 2, 3, 1, 3, 40],
                [3, 2, 6, 0, 4, 80],
                [3, 1, 6, 1, 2, 112],
                [5, 2, 6, 1, 3, 160],
                [3, 1, 6, 0, 1, 320],
            ]

        base_planes = round_to_multiple_of(base_planes * width_multiplier, round_nearest)
        # 参考Torchvision实现，不对输出特征维度进行缩放
        # out_planes = round_to_multiple_of(out_planes * width_multiplier, round_nearest)
        for i in range(len(stage_setting)):
            stage_setting[i][-1] = round_to_multiple_of(stage_setting[i][-1] * width_multiplier, round_nearest)

        self.first_stem, self.last_stem = make_stem(in_channels,
                                                    base_planes,
                                                    stage_setting[-1][-1],
                                                    out_channels,
                                                    conv_layer,
                                                    norm_layer,
                                                    act_layer)

        in_channels = base_planes
        for i, (kernel_size, stride, expansion_rate, with_attention_2, repeated, out_channels) in enumerate(
                stage_setting):
            features = list()
            for j in range(repeated):
                stride = stride if j == 0 else 1
                features.append(block_layer(in_channels,
                                            out_channels,
                                            stride=stride,
                                            kernel_size=kernel_size,
                                            expansion_rate=expansion_rate,
                                            with_attention=with_attention and with_attention_2,
                                            reduction=reduction,
                                            attention_type=attention_type,
                                            conv_layer=conv_layer,
                                            norm_layer=norm_layer,
                                            act_layer=act_layer
                                            ))
                in_channels = out_channels
            stage_name = f'stage{i + 1}'
            self.add_module(stage_name, nn.Sequential(*features))
        self.stage_num = len(stage_setting)

        self._init_weights()

    def _init_weights(self):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                nn.init.kaiming_uniform_(m.weight, mode="fan_out", nonlinearity="sigmoid")
                if m.bias is not None:
                    nn.init.zeros_(m.bias)

    def forward(self, x):
        x = self.first_stem(x)

        for i in range(self.stage_num):
            stage = self.__getattr__(f'stage{i + 1}')
            x = stage(x)

        x = self.last_stem(x)
        return x


@registry.Backbone.register('MNASNet')
def build_mnasnet(cfg):
    arch = cfg.MODEL.BACKBONE.ARCH
    in_planes = cfg.MODEL.BACKBONE.IN_PLANES
    norm_layer = get_norm(cfg)
    # compression
    width_multiplier = cfg.MODEL.COMPRESSION.WIDTH_MULTIPLIER
    round_nearest = cfg.MODEL.COMPRESSION.ROUND_NEAREST
    # attention
    with_attention = cfg.MODEL.ATTENTION.WITH_ATTENTION
    reduction = cfg.MODEL.ATTENTION.REDUCTION
    attention_type = cfg.MODEL.ATTENTION.ATTENTION_TYPE
    # conv
    conv_layer = get_conv(cfg)
    # act
    act_layer = get_act(cfg)

    out_planes = 1280
    stage_setting = copy.deepcopy(arch_settings[arch])
    return MNASNetBackbone(
        in_channels=in_planes,
        out_channels=out_planes,
        stage_setting=stage_setting,
        width_multiplier=width_multiplier,
        round_nearest=round_nearest,
        with_attention=with_attention,
        reduction=reduction,
        attention_type=attention_type,
        conv_layer=conv_layer,
        norm_layer=norm_layer,
        act_layer=act_layer
    )



# -*- coding: utf-8 -*-

"""
@date: 2020/8/22 下午9:02
@file: adam.py
@author: zj
@description: 
"""

import torch.nn as nn
import torch.optim as optim

from .. import registry


@registry.OPTIMIZERS.register('ADAM')
def build_adam(cfg, groups):
    lr = cfg.OPTIMIZER.LR
    weight_decay = cfg.OPTIMIZER.WEIGHT_DECAY.DECAY

    return optim.Adam(groups, lr=lr, weight_decay=weight_decay)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import importlib
import numpy as np
import os
import re
import subprocess
import sys
from collections import defaultdict
import PIL
import torch
import torchvision
from tabulate import tabulate

__all__ = ["collect_env_info"]


def collect_torch_env():
    try:
        import torch.__config__

        return torch.__config__.show()
    except ImportError:
        # compatible with older versions of pytorch
        from torch.utils.collect_env import get_pretty_env_info

        return get_pretty_env_info()


def get_env_module():
    var_name = "DETECTRON2_ENV_MODULE"
    return var_name, os.environ.get(var_name, "<not set>")


def detect_compute_compatibility(CUDA_HOME, so_file):
    try:
        cuobjdump = os.path.join(CUDA_HOME, "bin", "cuobjdump")
        if os.path.isfile(cuobjdump):
            output = subprocess.check_output(
                "'{}' --list-elf '{}'".format(cuobjdump, so_file), shell=True
            )
            output = output.decode("utf-8").strip().split("\n")
            sm = []
            for line in output:
                line = re.findall(r"\.sm_[0-9]*\.", line)[0]
                sm.append(line.strip("."))
            sm = sorted(set(sm))
            return ", ".join(sm)
        else:
            return so_file + "; cannot find cuobjdump"
    except Exception:
        # unhandled failure
        return so_file


def collect_env_info():
    has_gpu = torch.cuda.is_available()  # true for both CUDA & ROCM
    torch_version = torch.__version__

    # NOTE: the use of CUDA_HOME and ROCM_HOME requires the CUDA/ROCM build deps, though in
    # theory detectron2 should be made runnable with only the corresponding runtimes
    from torch.utils.cpp_extension import CUDA_HOME

    has_rocm = False
    if tuple(map(int, torch_version.split(".")[:2])) >= (1, 5):
        from torch.utils.cpp_extension import ROCM_HOME

        if (getattr(torch.version, "hip", None) is not None) and (ROCM_HOME is not None):
            has_rocm = True
    has_cuda = has_gpu and (not has_rocm)

    data = []
    data.append(("sys.platform", sys.platform))
    data.append(("Python", sys.version.replace("\n", "")))
    data.append(("numpy", np.__version__))

    # try:
    #     import detectron2  # noqa
    #
    #     data.append(
    #         ("detectron2", detectron2.__version__ + " @" + os.path.dirname(detectron2.__file__))
    #     )
    # except ImportError:
    #     data.append(("detectron2", "failed to import"))
    #
    # try:
    #     from detectron2 import _C
    # except ImportError:
    #     data.append(("detectron2._C", "failed to import"))
    #
    #     # print system compilers when extension fails to build
    #     if sys.platform != "win32":  # don't know what to do for windows
    #         try:
    #             # this is how torch/utils/cpp_extensions.py choose compiler
    #             cxx = os.environ.get("CXX", "c++")
    #             cxx = subprocess.check_output("'{}' --version".format(cxx), shell=True)
    #             cxx = cxx.decode("utf-8").strip().split("\n")[0]
    #         except subprocess.SubprocessError:
    #             cxx = "Not found"
    #         data.append(("Compiler", cxx))
    #
    #         if has_cuda and CUDA_HOME is not None:
    #             try:
    #                 nvcc = os.path.join(CUDA_HOME, "bin", "nvcc")
    #                 nvcc = subprocess.check_output("'{}' -V".format(nvcc), shell=True)
    #                 nvcc = nvcc.decode("utf-8").strip().split("\n")[-1]
    #             except subprocess.SubprocessError:
    #                 nvcc = "Not found"
    #             data.append(("CUDA compiler", nvcc))
    # else:
    #     # print compilers that are used to build extension
    #     data.append(("Compiler", _C.get_compiler_version()))
    #     data.append(("CUDA compiler", _C.get_cuda_version()))  # cuda or hip
    #     if has_cuda:
    #         data.append(
    #             ("detectron2 arch flags", detect_compute_compatibility(CUDA_HOME, _C.__file__))
    #         )
    #
    # data.append(get_env_module())
    data.append(("PyTorch", torch_version + " @" + os.path.dirname(torch.__file__)))
    data.append(("PyTorch debug build", torch.version.debug))

    data.append(("GPU available", has_gpu))
    if has_gpu:
        devices = defaultdict(list)
        for k in range(torch.cuda.device_count()):
            devices[torch.cuda.get_device_name(k)].append(str(k))
        for name, devids in devices.items():
            data.append(("GPU " + ",".join(devids), name))

        if has_rocm:
            data.append(("ROCM_HOME", str(ROCM_HOME)))
        else:
            data.append(("CUDA_HOME", str(CUDA_HOME)))

            cuda_arch_list = os.environ.get("TORCH_CUDA_ARCH_LIST", None)
            if cuda_arch_list:
                data.append(("TORCH_CUDA_ARCH_LIST", cuda_arch_list))
    data.append(("Pillow", PIL.__version__))

    try:
        data.append(
            (
                "torchvision",
                str(torchvision.__version__) + " @" + os.path.dirname(torchvision.__file__),
            )
        )
        if has_cuda:
            try:
                torchvision_C = importlib.util.find_spec("torchvision._C").origin
                msg = detect_compute_compatibility(CUDA_HOME, torchvision_C)
                data.append(("torchvision arch flags", msg))
            except ImportError:
                data.append(("torchvision._C", "failed to find"))
    except AttributeError:
        data.append(("torchvision", "unknown"))

    # try:
    #     import fvcore
    #
    #     data.append(("fvcore", fvcore.__version__))
    # except ImportError:
    #     pass

    try:
        import cv2

        data.append(("cv2", cv2.__version__))
    except ImportError:
        pass
    env_str = tabulate(data) + "\n"
    env_str += collect_torch_env()
    return env_str


if __name__ == "__main__":
    # try:
    #     import detectron2  # noqa
    # except ImportError:
    #     print(collect_env_info())
    # else:
    #     from detectron2.utils.collect_env import collect_env_info

    print(collect_env_info())



# -*- coding: utf-8 -*-

"""
@date: 2021/1/4 上午11:32
@file: split_attention_conv2d.py
@author: zj
@description: 
"""
from abc import ABC

import torch

import torch.nn as nn
import torch.nn.functional as F

from ..init_helper import init_weights


class SplitAttentionConv2d(nn.Module, ABC):

    def __init__(self,
                 in_channels,
                 out_channels,
                 radix=2,
                 groups=1,
                 reduction_rate=4,
                 default_channels: int = 32,
                 dimension: int = 2
                 ):
        """
        Implementation of SplitAttention in ResNetSt, refer to
        1. https://github.com/open-mmlab/mmdetection/blob/master/mmdet/models/backbones/resnest.py
        2. https://github.com/zhanghang1989/ResNeSt/blob/73b43ba63d1034dbf3e96b3010a8f2eb4cc3854f/resnest/torch/splat.py
        Partial reference ./selective_kernel_conv2d.py implementation
        :param in_channels:
        :param out_channels:
        :param radix:
        :param groups:
        :param reduction_rate:
        :param default_channels:
        :param dimension:
        """
        super(SplitAttentionConv2d, self).__init__()

        # split
        self.split = nn.Sequential(
            nn.Conv2d(in_channels, out_channels * radix, kernel_size=3, stride=1, padding=1, bias=False,
                      groups=groups * radix),
            nn.BatchNorm2d(out_channels * radix),
            nn.ReLU(inplace=True)
        )
        # self.conv1 = nn.Conv2d(in_channels, out_channels * radix, kernel_size=3, stride=1, padding=1, bias=False,
        #                        groups=groups * radix)
        # self.bn1 = nn.BatchNorm2d(out_channels * radix)
        # self.relu = nn.ReLU(inplace=True)
        # fuse
        self.pool = nn.AdaptiveAvgPool2d((1, 1))
        inner_channels = max(in_channels * radix // reduction_rate, 32)
        # inner_channels = max(out_channels // reduction_rate, default_channels)
        self.compact = nn.Sequential(
            nn.Conv2d(out_channels, inner_channels, kernel_size=1, stride=1, padding=0, bias=True,
                      groups=groups),
            nn.BatchNorm2d(inner_channels),
            nn.ReLU(inplace=True)
        )
        # select
        self.select = nn.Conv2d(inner_channels, out_channels * radix, kernel_size=1, stride=1,
                                groups=groups)
        # self.softmax = nn.Softmax(dim=0)
        self.rsoftmax = rSoftMax(radix, groups)
        self.dimension = dimension
        self.out_channels = out_channels
        self.radix = radix
        self.groups = groups

        init_weights(self.modules())

    def forward(self, x):
        N, C, H, W = x.shape[:4]
        # split
        out = self.split(x)
        # out = self.conv1(x)
        # out = self.bn1(out)
        # out = self.relu(out)
        split_out = torch.stack(torch.split(out, self.out_channels, dim=1))
        # fuse
        u = torch.sum(split_out, dim=0)
        s = self.pool(u)
        z = self.compact(s)
        # select
        c = self.select(z)
        softmax_c = self.rsoftmax(c).view(N, -1, 1, 1)

        attens = torch.split(softmax_c, self.out_channels, dim=1)
        v = sum([att * split for (att, split) in zip(attens, split_out)])
        return v.contiguous()
        # split_c = torch.stack(torch.split(c, self.out_channels, dim=1))
        # softmax_c = self.softmax(split_c)
        #
        # v = torch.sum(split_out.mul(softmax_c), dim=0)
        # return v.contiguous()


class rSoftMax(nn.Module):
    def __init__(self, radix, cardinality):
        super().__init__()
        self.radix = radix
        self.cardinality = cardinality

    def forward(self, x):
        batch = x.size(0)
        if self.radix > 1:
            x = x.view(batch, self.cardinality, self.radix, -1).transpose(1, 2)
            x = F.softmax(x, dim=1)
            x = x.reshape(batch, -1)
        else:
            x = torch.sigmoid(x)
        return x



# -*- coding: utf-8 -*-

"""
@date: 2021/7/23 下午9:27
@file: square_pad.py
@author: zj
@description: 
"""

import cv2
import albumentations as A


class Rotate(object):
    """
    Rotate the input by an angle selected randomly from the uniform distribution.

    Args:
        limit ((int, int) or int): range from which a random angle is picked. If limit is a single int
            an angle is picked from (-limit, limit). Default: (-90, 90)
        border_mode (OpenCV flag): flag that is used to specify the pixel extrapolation method. Should be one of:
            cv2.BORDER_CONSTANT, cv2.BORDER_REPLICATE, cv2.BORDER_REFLECT, cv2.BORDER_WRAP, cv2.BORDER_REFLECT_101.
            Default: cv2.BORDER_REFLECT_101
        interpolation (OpenCV flag): flag that is used to specify the interpolation algorithm. Should be one of:
            cv2.INTER_NEAREST, cv2.INTER_LINEAR, cv2.INTER_CUBIC, cv2.INTER_AREA, cv2.INTER_LANCZOS4.
            Default: cv2.INTER_LINEAR.
        value (int, float, list of ints, list of float): padding value if border_mode is cv2.BORDER_CONSTANT.
        p (float): probability of applying the transform. Default: 0.5.
    """

    def __init__(self, limit, interpolation=cv2.INTER_LINEAR, border_mode=cv2.BORDER_REFLECT_101, value=None,
                 p=0.5):
        self.limit = limit
        self.interpolation = interpolation
        self.border_mode = border_mode
        self.value = value
        self.p = p
        self.t = A.Rotate(limit=self.limit, interpolation=self.interpolation,
                          border_mode=self.border_mode, value=self.value, p=self.p)

    def __call__(self, image):
        return self.t(image=image)['image']

    def __repr__(self):
        return self.__class__.__name__ + '(limit={0}, interpolation={1}, border_mode={2}, value={3}, p={4})'. \
            format(self.limit, self.interpolation, self.border_mode, self.value, self.p)



# -*- coding: utf-8 -*-

"""
@date: 2020/11/27 下午3:39
@file: group_norm_wrapper.py
@author: zj
@description: 
"""

import torch.nn as nn


class GroupNormWrapper(nn.GroupNorm):

    def __init__(self, num_channels: int, num_groups: int = 4, eps: float = 1e-5, affine: bool = True) -> None:
        super(GroupNormWrapper, self).__init__(num_groups, num_channels, eps, affine)

        self.num_features = num_channels



# -*- coding: utf-8 -*-

"""
@date: 2021/10/3 下午12:06
@file: color_jitter.py
@author: zj
@description: 
"""

import albumentations as A


class CoarseDropout(object):
    """
    CoarseDropout of the rectangular regions in the image.

    Args:
        max_holes (int): Maximum number of regions to zero out.
        max_height (int): Maximum height of the hole.
        max_width (int): Maximum width of the hole.
        min_holes (int): Minimum number of regions to zero out. If `None`,
            `min_holes` is be set to `max_holes`. Default: `None`.
        min_height (int): Minimum height of the hole. Default: None. If `None`,
            `min_height` is set to `max_height`. Default: `None`.
        min_width (int): Minimum width of the hole. If `None`, `min_height` is
            set to `max_width`. Default: `None`.
        fill_value (int, float, list of int, list of float): value for dropped pixels.

    Image types:
        uint8, float32

    Reference:
    |  https://arxiv.org/abs/1708.04552
    |  https://github.com/uoguelph-mlrg/Cutout/blob/master/util/cutout.py
    |  https://github.com/aleju/imgaug/blob/master/imgaug/augmenters/arithmetic.py
    """

    def __init__(self, max_holes=8, max_height=8, max_width=8, min_holes=None, min_height=None, min_width=None,
                 fill_value=0, p=0.5):
        self.max_holes = max_holes
        self.max_height = max_height
        self.max_width = max_width
        self.min_holes = min_holes
        self.min_height = min_height
        self.min_width = min_width
        self.fill_value = fill_value
        self.p = p

        self.t = A.CoarseDropout(max_holes=self.max_holes, max_height=self.max_height, max_width=self.max_width,
                                 min_holes=self.min_holes, min_height=self.min_height, min_width=self.min_width,
                                 fill_value=self.fill_value, p=self.p)

    def __call__(self, image):
        return self.t(image=image)['image']

    def __repr__(self):
        return self.__class__.__name__ + '(max_holes={0}, max_height={1}, max_width={2}, ' \
                                         'min_holes={3}, min_height={4}, min_width={5}, ' \
                                         'fill_value={6}, p={7})' \
            .format(self.max_holes, self.max_height, self.max_width,
                    self.min_holes, self.min_height, self.min_width, self.fill_value, self.p)



# -*- coding: utf-8 -*-

"""
@date: 2021/9/27 下午7:43
@file: mp_dataset.py
@author: zj
@description: see [zjykzj/MPDataset](https://github.com/zjykzj/MPDataset)
"""

import os
import json

import torch
from torch.utils.data import IterableDataset
from torch.utils.data import RandomSampler, SequentialSampler
from torchvision.datasets.folder import default_loader

from zcls.config.key_word import KEY_IMGS, KEY_TARGETS, KEY_CLASSES
from ..samplers.distributed_sampler import DistributedSampler
from .evaluator.general_evaluator import GeneralEvaluator


def get_base_info(json_path):
    assert os.path.isfile(json_path), json_path
    with open(json_path, 'r') as f:
        data_dict = json.load(f)

    classes = data_dict[KEY_CLASSES]
    length = len(data_dict[KEY_TARGETS])

    return classes, length


def build_sampler(dataset, num_gpus=1, random_sample=False, rank_id=0, drop_last=False):
    if num_gpus <= 1:
        if random_sample:
            # different work use same generator
            generator = torch.Generator()
            generator.manual_seed(int(torch.empty((), dtype=torch.int64).random_().item()))
            sampler = RandomSampler(dataset, generator=generator)
        else:
            sampler = SequentialSampler(dataset)
    else:
        shuffle = random_sample
        # using dataset.length replace dataset
        sampler = DistributedSampler(dataset.length,
                                     num_replicas=num_gpus,
                                     rank=rank_id,
                                     shuffle=shuffle,
                                     drop_last=drop_last)

    return sampler


def get_total_data(json_path, classes):
    assert os.path.isfile(json_path)
    assert isinstance(classes, list)

    with open(json_path, 'r') as f:
        data_dict = json.load(f)

    total_img_list = data_dict[KEY_IMGS]
    total_label_list = data_dict[KEY_TARGETS]

    return total_img_list, total_label_list


def get_subset_data(json_path, classes, indices):
    total_img_list, total_label_list = get_total_data(json_path, classes)

    sub_img_list = [total_img_list[idx] for idx in indices]
    sub_label_list = [total_label_list[idx] for idx in indices]

    return sub_img_list, sub_label_list


def shuffle_dataset(sampler, cur_epoch, is_shuffle=False):
    """"
    Shuffles the data.
    Args:
        sampler (sampler): sampler to perform shuffle.
        cur_epoch (int): number of the current epoch.
        is_shuffle (bool): need to shuffle the data
    """
    if not is_shuffle:
        return
    assert isinstance(
        sampler, (RandomSampler, DistributedSampler)
    ), "Sampler type '{}' not supported".format(type(sampler))

    # RandomSampler handles shuffling automatically
    if isinstance(sampler, DistributedSampler):
        # DistributedSampler shuffles data based on epoch
        sampler.set_epoch(cur_epoch)


class MPDataset(IterableDataset):

    def __init__(self, root, transform=None, target_transform=None, top_k=(1, 5), shuffle: bool = False,
                 num_gpus: int = 1, rank_id: int = 0, epoch: int = 0, drop_last: bool = False):
        super(MPDataset).__init__()
        self.root = root
        self.transform = transform
        self.target_transform = target_transform
        self.shuffle = shuffle

        self.rank = rank_id
        self.num_replicas = num_gpus

        self.classes, self.length = get_base_info(root)
        self.sampler = build_sampler(self, num_gpus=self.num_replicas, random_sample=self.shuffle,
                                     rank_id=self.rank, drop_last=drop_last)
        self.set_epoch(epoch)

        self._update_evaluator(top_k)

    def parse_file(self, img_list, label_list):
        for img_path, target in zip(img_list, label_list):
            image = default_loader(img_path)
            if self.transform is not None:
                image = self.transform(image)
            if self.target_transform is not None:
                target = self.target_transform(target)

            yield image, target

    def get_indices(self):
        worker_info = torch.utils.data.get_worker_info()

        if worker_info is not None:
            # in a worker process
            worker_id = worker_info.id
            # split workload
            indices = self.indices[worker_id:self.indices_length:worker_info.num_workers]
        else:
            # single-process data loading, return the full iterator
            indices = self.indices

        return indices

    def __iter__(self):
        indices = self.get_indices()
        img_list, label_list = get_subset_data(self.root, self.classes, indices)
        assert len(img_list) == len(label_list)

        return iter(self.parse_file(img_list, label_list))

    def __len__(self):
        return self.indices_length if self.num_replicas > 1 else self.length

    def _update_evaluator(self, top_k):
        self.evaluator = GeneralEvaluator(self.classes, top_k=top_k)

    def set_epoch(self, epoch: int) -> None:
        r"""
        Sets the epoch for this sampler.

        Args:
            epoch (int): Epoch number.
        """
        shuffle_dataset(self.sampler, epoch, self.shuffle)
        self.indices = list(self.sampler)
        self.indices_length = len(self.indices)



# -*- coding: utf-8 -*-

"""
@date: 2021/10/3 下午1:21
@file: to_tensor.py
@author: zj
@description: 
"""

from albumentations.pytorch.transforms import ToTensorV2


class ToTensor(object):
    """
    Convert image to `torch.Tensor`. The numpy `HWC` image is converted to pytorch `CHW` tensor.
    If the image is in `HW` format (grayscale image), it will be converted to pytorch `HW` tensor.
    """

    def __init__(self, p=1.0):
        self.p = p

        self.t = ToTensorV2(p=self.p)

    def __call__(self, image):
        return self.t(image=image)['image']

    def __repr__(self):
        return self.__class__.__name__ + '(p={0})'.format(self.p)



# -*- coding: utf-8 -*-

"""
@date: 2020/11/25 下午6:51
@file: transform.py
@author: zj
@description: 
"""

from yacs.config import CfgNode as CN


def add_config(_C):
    # ---------------------------------------------------------------------------- #
    # Transform
    # ---------------------------------------------------------------------------- #
    _C.TRANSFORM = CN()
    _C.TRANSFORM.TRAIN_METHODS = ('Resize', 'CenterCrop', 'ToTensor', 'Normalize')
    _C.TRANSFORM.TEST_METHODS = ('Resize', 'CenterCrop', 'ToTensor', 'Normalize')

    # default: policy: AutoAugmentPolicy = AutoAugmentPolicy.IMAGENET, p=0.5
    # for policy, should be
    #     IMAGENET = "imagenet"
    #     CIFAR10 = "cifar10"
    #     SVHN = "svhn"
    _C.TRANSFORM.AUTOAUGMENT = ("imagenet", 0.5)

    # default: max_holes=8, max_height=8, max_width=8, min_holes=None, min_height=None, min_width=None, fill_value=0, p=0.5
    _C.TRANSFORM.COARSE_DROPOUT = (8, 8, 8, None, None, None, 0, 0.5)

    # default: brightness=0.2, contrast=0.2, saturation=0.2, hue=0.2, p=0.5
    _C.TRANSFORM.COLOR_JITTER = (0.2, 0.2, 0.2, 0.2, 0.5)

    # default: size, p=1.0
    _C.TRANSFORM.TRAIN_CENTER_CROP = ((224, 224), 1.0)
    _C.TRANSFORM.TEST_CENTER_CROP = ((224, 224), 1.0)

    # default: size, p=1.0
    _C.TRANSFORM.RANDOM_CROP = ((224, 224), 1.0)

    # default: p=0.5
    _C.TRANSFORM.HORIZONTAL_FLIP = 0.5

    # default: p=0.5
    _C.TRANSFORM.VERTICAL_FLIP = 0.5

    # default: size, interpolation=cv2.INTER_LINEAR, p=1.0
    # for interpolation, should be
    #     INTER_AREA = 3
    #     INTER_CUBIC = 2
    #     INTER_LANCZOS4 = 4
    #     INTER_LINEAR = 1
    #     INTER_NEAREST = 0
    _C.TRANSFORM.TRAIN_RESIZE = ((224, 224), 1, 1.0)
    _C.TRANSFORM.TEST_RESIZE = ((224, 224), 1, 1.0)

    # default: limit, interpolation=cv2.INTER_LINEAR, border_mode=cv2.BORDER_REFLECT_101, value=None, p=0.5
    # for interpolation, should be
    #     INTER_AREA = 3
    #     INTER_CUBIC = 2
    #     INTER_LANCZOS4 = 4
    #     INTER_LINEAR = 1
    #     INTER_NEAREST = 0
    # for padding_mode, should be
    #     BORDER_CONSTANT = 0
    #     BORDER_DEFAULT = 4
    #     BORDER_ISOLATED = 16
    #     BORDER_REFLECT = 2
    #     BORDER_REFLECT101 = 4
    _C.TRANSFORM.ROTATE = ((-30, 30), 1, 4, None, 0.5)

    # default: padding_position=A.PadIfNeeded.PositionType.CENTER, padding_mode=cv2.BORDER_CONSTANT, fill=0, p=1.0
    # for padding_position, should be
    #     CENTER = "center"
    #     TOP_LEFT = "top_left"
    #     TOP_RIGHT = "top_right"
    #     BOTTOM_LEFT = "bottom_left"
    #     BOTTOM_RIGHT = "bottom_right"
    # for padding_mode, should be
    #     BORDER_CONSTANT = 0
    #     BORDER_DEFAULT = 4
    #     BORDER_ISOLATED = 16
    #     BORDER_REFLECT = 2
    #     BORDER_REFLECT101 = 4
    _C.TRANSFORM.SQUARE_PAD = ("center", 0, 0, 1.0)

    # default: mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225), max_pixel_value=255.0, p=1.0
    # Normalization is applied by the formula: `img = (img - mean * max_pixel_value) / (std * max_pixel_value)`
    _C.TRANSFORM.NORMALIZE = ((0.445, 0.445, 0.445), (0.225, 0.225, 0.225), 255.0, 1.0)

    # default: p=1.0
    # Convert image to `torch.Tensor`. The numpy `HWC` image is converted to pytorch `CHW` tensor.
    _C.TRANSFORM.TO_TENSOR = 1.0



# -*- coding: utf-8 -*-

"""
@date: 2021/2/1 下午7:10
@file: asymmetric_convolution_block.py
@author: zj
@description: 
"""

import torch.nn as nn


class AsymmetricConvolutionBlock(nn.Module):
    """
    refer to [ACNet/acnet/acb.py](https://github.com/DingXiaoH/ACNet/blob/master/acnet/acb.py)
    """

    def __init__(self, in_channels, out_channels, kernel_size, stride=1, padding=0, dilation=1, groups=1,
                 padding_mode='zeros', use_affine=True, use_last_bn=False):
        super(AsymmetricConvolutionBlock, self).__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.dilation = dilation
        self.groups = groups
        self.padding_mode = padding_mode

        self.square_conv = nn.Conv2d(in_channels=in_channels, out_channels=out_channels,
                                     kernel_size=(kernel_size, kernel_size), stride=stride,
                                     padding=padding, dilation=dilation, groups=groups, bias=False,
                                     padding_mode=padding_mode)
        self.square_bn = nn.BatchNorm2d(num_features=out_channels, affine=use_affine)

        center_offset_from_origin_border = padding - kernel_size // 2
        ver_pad_or_crop = (padding, center_offset_from_origin_border)
        hor_pad_or_crop = (center_offset_from_origin_border, padding)
        if center_offset_from_origin_border >= 0:
            # padding
            self.ver_conv_crop_layer = nn.Identity()
            ver_conv_padding = ver_pad_or_crop
            self.hor_conv_crop_layer = nn.Identity()
            hor_conv_padding = hor_pad_or_crop
        else:
            # crop
            self.ver_conv_crop_layer = CropLayer(crop_set=ver_pad_or_crop)
            ver_conv_padding = (0, 0)
            self.hor_conv_crop_layer = CropLayer(crop_set=hor_pad_or_crop)
            hor_conv_padding = (0, 0)
        self.ver_conv = nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=(kernel_size, 1),
                                  stride=stride,
                                  padding=ver_conv_padding, dilation=dilation, groups=groups, bias=False,
                                  padding_mode=padding_mode)

        self.hor_conv = nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=(1, kernel_size),
                                  stride=stride,
                                  padding=hor_conv_padding, dilation=dilation, groups=groups, bias=False,
                                  padding_mode=padding_mode)
        self.ver_bn = nn.BatchNorm2d(num_features=out_channels, affine=use_affine)
        self.hor_bn = nn.BatchNorm2d(num_features=out_channels, affine=use_affine)

        if use_last_bn:
            self.last_bn = nn.BatchNorm2d(num_features=out_channels, affine=True)

        self._init_weights()

    def _init_weights(self, gamma=1.0 / 3):
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, gamma)
                nn.init.constant_(m.bias, gamma)

    def forward(self, input):
        square_outputs = self.square_conv(input)
        square_outputs = self.square_bn(square_outputs)
        vertical_outputs = self.ver_conv_crop_layer(input)
        vertical_outputs = self.ver_conv(vertical_outputs)
        vertical_outputs = self.ver_bn(vertical_outputs)
        horizontal_outputs = self.hor_conv_crop_layer(input)
        horizontal_outputs = self.hor_conv(horizontal_outputs)
        horizontal_outputs = self.hor_bn(horizontal_outputs)
        result = square_outputs + vertical_outputs + horizontal_outputs
        if hasattr(self, 'last_bn'):
            return self.last_bn(result)
        return result


class CropLayer(nn.Module):

    #   E.g., (-1, 0) means this layer should crop the first and last rows of the feature map. And (0, -1) crops the first and last columns
    def __init__(self, crop_set):
        super(CropLayer, self).__init__()
        self.rows_to_crop = - crop_set[0]
        self.cols_to_crop = - crop_set[1]
        assert self.rows_to_crop >= 0
        assert self.cols_to_crop >= 0

    def forward(self, input):
        if self.rows_to_crop == 0 and self.cols_to_crop == 0:
            return input
        elif self.rows_to_crop > 0 and self.cols_to_crop == 0:
            return input[:, :, self.rows_to_crop:-self.rows_to_crop, :]
        elif self.rows_to_crop == 0 and self.cols_to_crop > 0:
            return input[:, :, :, self.cols_to_crop:-self.cols_to_crop]
        else:
            return input[:, :, self.rows_to_crop:-self.rows_to_crop, self.cols_to_crop:-self.cols_to_crop]



# -*- coding: utf-8 -*-

"""
@date: 2021/10/3 下午1:25
@file: __init__.py.py
@author: zj
@description: 
"""

__all__ = ['AutoAugment', 'CoarseDropout', 'ColorJitter', 'HorizontalFlip', 'VerticalFlip', 'RandomCrop', 'CenterCrop',
           'Resize', 'Rotate', 'SquarePad', 'Normalize', 'ToTensor']

from .autoaugment import AutoAugment

from .coarse_dropout import CoarseDropout
from .color_jitter import ColorJitter

from .horizontal_flip import HorizontalFlip
from .vertical_flip import VerticalFlip

from .random_crop import RandomCrop
from .center_crop import CenterCrop

from .resize import Resize
from .rotate import Rotate
from .square_pad import SquarePad

from .normalize import Normalize
from .to_tensor import ToTensor
